:- use_module(library(terms), [ subsumes_chk/2 ]).
:- use_module(library(lists), [ append/3 ]).

%%% MOVE CLAUSE
move_clause(X,Y,Z) :- apply_move_clause(X,Y,Z).

apply_move_clause(error(Alert),CN,_) :- 
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.
	
apply_move_clause(error(Alert),_,CN2) :-
	(CN2=\=0, \+ is_in_current_prog(CN2)), !,
         Alert='New position undefined in current program.'.

apply_move_clause(Output,CN,CN2) :-
	Output=out(label(move_clause(CN,CN2)),[],[]).

%%% REPLACE CLAUSES
replace_clauses(X,Y) :- apply_replace_clauses(X,Y).

apply_replace_clauses(error(Alert),CNs) :-
	member(CN,CNs),
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.

apply_replace_clauses(Output,ToRep) :-
	parseNewClauses(map_tmp_newdef,NewPreds,Cls,Success),
	(Success == error -> Output=error('Syntax error in clauses.') ;
		 (Success == wait -> fail ;
			  (Msg1='{Warning:  the correctness of this transformation is not ensured by the system.',
					Msg2=' Be sure that it is semantics preserving.',
					Msg3='\nContinue replace clauses ?}',
					concat_atoms([Msg1,Msg2,Msg3],Msg),
					tcl_warning_yesno(Msg,Answer),
					(Answer=="yes" -> true ;
						 Output=error('Abort replace clauses.'), fail),
					Output=out(label(replace_clauses(NewPreds)),ToRep,Cls),
					tcl_eval('destroy .wrepclauses')))).

%%% ADD CLAUSES
add_clauses(X) :- apply_add_clauses(X).

apply_add_clauses(Output) :-
	parseNewClauses(map_tmp_newdef,NewPreds,Cls,Success),
	(Success == error -> Output=error('Syntax error in clauses.') ;
    	(Success == wait -> fail ;
	    (current_program(_,CNs),
		( check_add_clauses(CNs,Cls) -> true ;
		    (	Msg1='{Warning: Clause(s) to add are not subsumed by current clauses.',
			Msg2=' This might change the semantics of the program.',
			Msg3='\nAdd clauses anyway ?}',
			concat_atoms([Msg1,Msg2,Msg3],Msg),
			tcl_warning_yesno(Msg,Answer),
			( Answer=="yes" -> true ; Output=error('Abort add clauses.'), fail))),
		Output=out(label(add_clauses(NewPreds)),[],Cls),
		tcl_eval('destroy .wnewclauses')))).


check_add_clauses(_,[]).
check_add_clauses(CNs,[Cl|Cls]) :-
	is_a_tautology_cl(Cl), !,
	check_add_clauses(CNs,Cls).
check_add_clauses(CNs,[Cl|Cls]) :-
	member(CN2,CNs), clause_hd_bd(CN2,H2,bd([],B2)),
	is_subsumed_by(Cl,(H2,B2)),
	check_add_clauses(CNs,Cls).

%%% DELETE CLAUSES
delete_clauses(X,Y) :- apply_delete_clauses(X,Y).

apply_delete_clauses(error(Alert),CNs) :-
	member(CN,CNs),
	\+ is_in_current_prog(CN), !,
         Alert='Not in current program.'.


apply_delete_clauses(Output,ToDel) :-
	current_program(_,CNs),
	( check_delete_clauses(CNs,ToDel) -> true ;
	    (Msg1='{Warning: Deletion is not applicable to the selected clause(s).',
		Msg2=' (It applies only to subsumed clauses, clauses with a failing body, or tautologies).',
		Msg3='\nDelete clauses anyway ?}',
		concat_atoms([Msg1,Msg2,Msg3],Msg),
	    tcl_warning_yesno(Msg,Answer),
	    (Answer=="yes" -> true ; Output=error('Abort delete clauses.') ; fail))),
	Output=out(label(delete_clauses(ToDel)),ToDel,[]).


check_delete_clauses(_,[]).
check_delete_clauses(Cs,[C|ToDel]) :-
	is_a_tautology(C), !,
	check_delete_clauses(Cs,ToDel).
check_delete_clauses(Cs,[C|ToDel]) :-
	has_a_failing_body(C), !,
	check_delete_clauses(Cs,ToDel).
check_delete_clauses(Cs,[C|ToDel]) :-
	is_subsumed_in_prog(C,Cs),
	check_delete_clauses(Cs,ToDel).


is_subsumed_in_prog(CN,CNs) :-
	member(CN2,CNs),
	CN =\= CN2,
	clause_hd_bd(CN,H1,bd([],B1)),
	clause_hd_bd(CN2,H2,bd([],B2)),
	is_subsumed_by((H1,B1),(H2,B2)).

is_subsumed_by((H1,B1),(H2,B2)) :-
	(subsumes_chk( (H2,B2), (H1,B1))
	;   (append(F,_,B1), subsumes_chk((H2,B2), (H1,F)))).

is_a_tautology(CN) :-
	clause_hd_bd(CN,H1,bd([],B1)), is_a_tautology_cl((H1,B1)).
	
is_a_tautology_cl((H1,B1)) :- member(A,B1), H1==A.
is_a_tautology_cl((_,B1)) :- member(A,B1), member(NotA,B1),
	opposite_literals(A,NotA).

tcl_prepare_new_clauses(Where,L) :-
	get_clause_numbers(L,CNs),
	tcl_add_clauses(Where,CNs).

tcl_add_clauses(_,[]).
tcl_add_clauses(Where,[CN|L]) :-
	clause_hd_bd(CN,H,bd(Cs,As)),
	append(Cs,As,B),
	make_clause_readable_(H,B,Head,Body),
        concat_atoms(['addClause ',Where,' { ',Head,'} ',Body],Cmd),
	tcl_eval(Cmd),
	tcl_add_clauses(Where,L).

parseNewClauses(File,NewPreds,Cls,Success) :-
	see(File),
	parseNewCls(Cls,NewPreds,SingVars,UndefLits,Success),
	seen,
	(Success == error -> true ;
    	(Success == wait -> true ;
	    (flatten(SingVars,LSingVars),
	    flatten(UndefLits,LUndefLitsTmp2),
  	    list_to_ord_set(LUndefLitsTmp2,LUndefLitsTmp),
	    delete_elts_sublist(NewPreds,LUndefLitsTmp,LUndefLits),
	    tcl_warning_new_definition2(LSingVars,LUndefLits,Answer),
	    ( Answer=="no" -> Success=wait ; true)))).


parseNewCls(Cs2,NewPreds,Ss2,UndefLits2,Success) :-
	read_term(Cl,[singletons(S)]),
	(Cl==end_of_file ->
		 Cs2=[], NewPreds=[],Ss2=[],UndefLits2=[]
	;
		 ((Cl= (Head :- Body_Conj) ->
			conj_to_list(Body_Conj,Body),
		 %%% may be get_current_solver?			
			find_constraints(_Solver,Body,Constrs,Atoms),
			C=(Head,bd(Constrs,Atoms))			
        ;
			Cl=Head, Body=[], C=(Cl,bd([],[]))
		  ),
			  undefined_lits_in_body([Head|Body],U),
			  atom_pred_arity(Head,P,A),
			  parseNewCls2(Cs,[P/A],NewPreds,Ss,UndefLits,Success),
			  Cs2=[C|Cs], Ss2=[S|Ss], UndefLits2=[U|UndefLits]
		 )
	).
parseNewCls(_,_,_,_,error).
	
parseNewCls2([C|Cs],NewPredsTmp,NewPreds,[S|Ss],[U|UndefLits],Success) :-
    read_term(Cl,[singletons(S)]),
    Cl\==end_of_file, !,
    (Cl= (Head :- Body_Conj) ->
		  conj_to_list(Body_Conj,Body),
		 %%% may be get_current_solver?			
		  find_constraints(_Solver,Body,Constrs,Atoms),
		  C=(Head,bd(Constrs,Atoms))
	 ;
		  Cl=Head, Body=[], C=(Cl,bd([],[]))
	 ),
	 undefined_lits_in_body([Head|Body],U),
	 atom_pred_arity(Head,P,A),
	 insert_once(P/A,NewPredsTmp,NewPredsTmp2),
	 parseNewCls2(Cs,NewPredsTmp2,NewPreds,Ss,UndefLits,Success).
parseNewCls2([],L,L,[],[],_) :- !.
parseNewCls2(_,_,_,_,_,error) :- !.
    
    
delete_any_clauses(X,Y) :- apply_delete_any_clauses(X,Y).
apply_delete_any_clauses(Output,CNs) :-
	current_program(_,CurCNs),
	subtract(CNs,CurCNs,Del,Dif),
	(Dif==[] -> true ; tcl_warning_not_in_current_program(Dif)),
	Output=out(label(delete_any_clauses(Del)),Del,[]).

