:- use_module(library(lists), [  permutation/2, nth/4, same_length/2 ]).
:- use_module(library(ordsets), [ list_to_ord_set/2 ]).
:- use_module(library(terms), [ variant/2,
                                subsumes/2
                                 ]).


fold(X,Y,Z,T) :- apply_ts_fold(X,Y,Z,T).

apply_ts_fold(error(Alert),CN_to_fold,_,_) :-
	\+ is_in_current_prog(CN_to_fold), !,
         Alert='Not in current program.'.

apply_ts_fold(error(Alert),_,CN_to_use,_) :-
	\+ in_ancestor(CN_to_use), !,
	concat_atoms([CN_to_use,' is not in an ancestor of current program.'], Alert).

%Comment the following until the corresponding predicate is defined
apply_ts_fold(error(Alert),_,CN_to_use,_) :-
	\+ is_a_definition_clause(CN_to_use), !,
	concat_atoms([CN_to_use,' is not a definition clause.'], Alert).

		   
apply_ts_fold(Output,CN_to_fold,CN_to_use,PosAtoms) :-
	clause_hd_bd(CN_to_fold,Hd,bd([],Bd)),
	clause_hd_bd(CN_to_use,Hd_def,bd([],Bd_def)),
	list_to_ord_set(PosAtoms,OrdPosAtoms),
	ord_positions_to_elts(Bd,OrdPosAtoms,Atoms,Rest),
	ts_conditions(Hd_def,Bd_def,Hd,Atoms,Rest), !,
	OrdPosAtoms=[InsertPos|_],
        nth(InsertPos,NewBody,Hd_def,Rest),
	NewCl=(Hd,bd([],NewBody)),
	Output=out(label(fold(CN_to_fold,CN_to_use,PosAtoms)),
		   [CN_to_fold], [NewCl]).

apply_ts_fold(error(Alert),_,_,_) :-
	Alert='Impossible folding.'.

% ts_conditions(H_def,B_def,H,Atoms,Rest): let C and D be respectively two clauses such that
% head(C)=H, body(C) is made out of the atoms of the two lists Atoms Rest,
% head(D)=H_def and body(D) is made out of the atoms in list B_def.
% We check that (C,Atoms,D) satisfies the Tamaki-Sato conditions.
ts_conditions(H_def,B_Def,H,Atoms,Rest) :-
    int_var(H_def,B_Def,V), subsumes(B_Def,Atoms), 
    free_variables(H_def,[],[],V1), intersection_vars_empty(V,V1),
    free_variables([H|Rest],[],[],V2), intersection_vars_empty(V,V2).





general_fold(Out,ToFold,ToUse) :-  
   check_input_clauses_for_folding(Out,ToFold,ToUse,GenH,A),
   (var(Out) -> apply_general_fold(Out,ToFold,ToUse,GenH,A,_) ;
       true).


fold(Out,ToFold,ToUse) :- 
	check_input_clauses_for_folding(Out,ToFold,ToUse,GenH,A),
	( var(Out) -> apply_fold(Out,ToFold,ToUse,GenH,A,_) ;
	    true).

check_input_clauses_for_folding(Out,ToFold,ToUse,GenH,A) :-
	current_program(PN,Cls),
	(subset(ToFold,Cls) ->
	    ( same_definition_clauses(ToUse,Anc,Out) ->
		( var(Out) ->
		    ( is_an_ancestor(Anc,PN) ->
			(   heads_unifiable(ToFold,_,GenH),
			    heads_unifiable(ToUse,_,A) ,
			    ( no_other_head_definition_unifiable(ToUse,A) -> true ;
				Out=error('The selected definition clauses do not form a definition: there exists at least one non selected definition clause whose head is unifiable with the head of the selected ones') ) ) ;
			Out=error('Definition clauses must belong to an ancestor of current program')) ) ;
			Out=error('Clauses used for folding must belong to a same definition.')) ;
	    Out=error('Clauses to fold must belong to current program.')).

same_definition_clauses([C|ToUse],Anc,Out) :-
	is_a_definition_clause(C,Pred),
	same_definition_clauses(ToUse,Pred,Anc,Out).
same_definition_clauses([C|ToUse],Pred,Anc,Out) :-
	is_a_definition_clause(C,Pred),
	same_definition_clauses(ToUse,Pred,Anc,Out).
same_definition_clauses([],_,_,_).

no_other_head_definition_unifiable(ToUse,A) :-
	bagof( CN, is_a_definition_clause(CN), Defs),
	subtract(Defs,ToUse,_,Dif),
	member(C,Dif),
	clause_hd_bd(C,Head,_),
	unify_with_occurs_check(Head,A), !, fail.
no_other_head_definition_unifiable(_,_).

apply_general_fold(Out,CsN,CsN2,GenHead,A,NewClause) :- 
   (general_fold_cond2(CsN,CsN2,GenHead,A,NewClause) ->
       Out=out(label(fold(CsN,CsN2)),CsN,[NewClause]) ;
       Out=error('Impossible folding.')).

apply_fold(Out,CsN,CsN2,GenHead,A,NewClause) :- 
   (fold_cond2(CsN,CsN2,GenHead,A,NewClause) ->
       Out=out(label(fold(CsN,CsN2)),CsN,[NewClause]) ;
       Out=error('Impossible folding.')).






/*-------------
 fold_cond2(Clauses, Defs, Gen_Head, A, Folded) is true if Folded is the clause
 which corresponds to the folding of the list Clauses of clauses wrt the list
 Defs of clauses. Depending on the use of fold_cond2 (within an interactive
 folding step with parameters Clauses and Defs selected by the user or in the
 context of the strategy determ, the checking specified by predicate
 check_input_clauses_for_folding in order to fold Clauses wrt Defs has been
 performed or not. If it has been performed (interactive folding step),
 Gen_Head and A are known (non variable terms). Otherwise, they are inferred by
 using predicate "head_unifiable" (the other conditions specified in the file
 header are always satisfied when a folding step is attempted in the "determ"
 strategy. Since the presence of unit clauses create some problems, they are
 first put aside (omit_first_unit_clauses), then the folding is attempted on
 the non unit clauses (fold_cond2_), and if it succeeds, it is verified that it
 is correct also for the unit clauses (verify_recover_unit_clauses).
--------------*/


fold_cond2(Clauses, Defs, Gen_Head, A, Folded):-
   omit_first_unit_clauses(Clauses,Defs,Units,DefUnits,Clauses2,Defs2),
   permutation(Clauses2,PClauses2),
   fold_cond2_(PClauses2, Defs2, Gen_Head, A, Folded),
   verify_recover_unit_clauses(Folded,Units,DefUnits).

fold_cond2_(Cls, Defs, Gen_Head, A, Folded):-
    (var(Gen_Head) -> heads_unifiable(Cls,_,Gen_Head) ; true),
    (var(A) -> heads_unifiable(Defs,_,A) ; true),
    minimal_hd_generalization_clauses(Cls,Gen_Head,_,_,Gen_Cl),
    minimal_hd_generalization_clauses(Defs,A,_,_,Gen_Def),
    recover_folded_clause(Gen_Cl,Gen_Def, Folded).

/*------------
 general_fold_cond2 is a generalization of fold_cond2 where Clauses might be
 reordered to perform the folding step.
-------------*/

general_fold_cond2(Clauses, Defs, Gen_Head, A, Folded):-
   omit_first_unit_clauses(Clauses,Defs,Units,DefUnits,Clauses2,Defs2),
   permutation(Clauses2,PClauses2),
   general_fold_cond2_(PClauses2, Defs2, Gen_Head, A, Folded),
   verify_recover_unit_clauses(Folded,Units,DefUnits).

general_fold_cond2_(Clauses, Defs, Gen_Head, A, Folded):-
	(var(Gen_Head) -> heads_unifiable(Clauses,_,Gen_Head) ; true),
	(var(A) -> heads_unifiable(Defs,_,A) ; true),
	minimal_hd_generalization_clauses(Clauses,Gen_Head,_,_,Gen_Clause),
	minimal_hd_generalization_clauses(Defs,A,_,_,Gen_Def),
	general_recover_folded_clause(Gen_Clause,Gen_Def, Folded).


%%% minimal_hd_generalization_clauses(NumCls,Gen_Head,Cls,Gen_Cls,Dis_Cl): NumCls is
%%% a list of clause numbers corresponding to clauses Cls, Gen_Head is an atom which
%%% generalizes the head of each clause in Cls, Gen_Cls is the list Cls of clauses
%%% after generalization so that their heads is Gen_Head (they might contain
%%% equalities in their bodies) Dis_Cl is a disjunctive clause with head Gen_Head and
%%% whose body is the disjunction of the bodies in Gen_Cls

minimal_hd_generalization_clauses(NumCls,Gen_Head,Cls,Gen_Cls,Dis_Cl) :-
    Gen_Head=..[_|Args_Gen_Head],
    minimal_hd_generalization_clauses(NumCls,Args_Gen_Head,Cls,Gen_Cls),
    disjunctive_clause(Gen_Cls,Dis_Cl).

minimal_hd_generalization_clauses([],_,[],[]).
minimal_hd_generalization_clauses([N|P],Gen_Head,[(Hd,bd([],Bd))|L],[Cl|Cls]) :-
    clause_hd_bd(N,Hd,bd(C,Bd)), Hd=..[F|Args],
    minimal_move_arguments_in_body(Args,Gen_Head,GenArgs,Eqs),
    GenHd=..[F|GenArgs], 
    insere_equalities_after_inequalities(Eqs,Bd,ExtBody),
    Cl=(GenHd,bd(C,ExtBody)),
    minimal_hd_generalization_clauses(P,Gen_Head,L,Cls).


minimal_hd_generalization_clauses_([],_,[],[]).
minimal_hd_generalization_clauses_([(Hd,bd(C,Bd))|P],Gen_Head,[(Hd,bd([],Bd))|L],[Cl|Cls]) :-
    Hd=..[F|Args],
    minimal_move_arguments_in_body(Args,Gen_Head,GenArgs,Eqs),
    GenHd=..[F|GenArgs], 
    insere_equalities_after_inequalities(Eqs,Bd,ExtBody),
    Cl=(GenHd,bd(C,ExtBody)),
    minimal_hd_generalization_clauses_(P,Gen_Head,L,Cls).

disjunctive_clause([(H,Bd)], (H, [Bd]) ) :-!.
disjunctive_clause([(H,Bd)|Gen_Clauses], (H, [Bd | Bds]) ) :-
    disjunctive_clause(Gen_Clauses, (H, Bds)).

minimal_move_arguments_in_body([],_,[],[]).
minimal_move_arguments_in_body([A|Args],[B|GenHead],[A|GenArgs],Eqs) :-
     var(A), occurs_at_same_pos(A,Args,B,GenHead), !, A = B,
     minimal_move_arguments_in_body(Args,GenHead,GenArgs,Eqs).
minimal_move_arguments_in_body([A|Args],[B|GenHead],[X|GenArgs],[X=A|Eqs]) :-
    var(B), !, minimal_move_arguments_in_body(Args,GenHead,GenArgs,Eqs).
minimal_move_arguments_in_body([A|Args],[B|GenHead],[X|GenArgs],Eqs) :-
    A =.. [F|ArgsA], B =.. [F|ArgsB], copy_term(B,X),
    X =.. [F|ArgsX], append(ArgsA,Args,ArgsA2),
    append(ArgsB,GenHead,GenHead2), append(ArgsX,GenArgs,GenArgs2),
    minimal_move_arguments_in_body(ArgsA2,GenHead2,GenArgs2,Eqs).



% Particular case where Args2 is a term more general than Args1
occurs_at_same_pos(V1,Args1,_,_) :- free_of_var(V1,Args1), !.
occurs_at_same_pos(_,[],_,[]).
occurs_at_same_pos(A,[T1|T1s],B,[T2|T2s]) :-
    occurs_at_same_pos_in_term(A,T1,B,T2), occurs_at_same_pos(A,T1s,B,T2s).
occurs_at_same_pos_in_term(A,T1,_,_) :- free_of_var(A,T1), !.
occurs_at_same_pos_in_term(A,T1,B,T2) :-  A == T1, B == T2.
occurs_at_same_pos_in_term(A,T1,B,T2) :-
    nonvar(T1), nonvar(T2), T1=..[_|Args1], T2=..[_|Args2],
    occurs_at_same_pos(A,Args1,B,Args2).

insere_equalities_after_inequalities(Body,[X|B2],[X|ExtBody]):-
    (var(X) ; nonvar(X), X = \+(_=_)), !,
    insere_equalities_after_inequalities(Body,B2,ExtBody).
insere_equalities_after_inequalities([],B,B) :- !.
insere_equalities_after_inequalities(Body,B2,B) :- append(Body,B2,B).

% general_recover_folded_clause/3 performs main folding
% (H1,Bds) clause to fold (generalized)
% (A,Bds2) definition to use (generalized)
% (H1,New_Bd_C) derived clause (possibly disjunctive)

general_recover_folded_clause( (H1,Bds), (A,Bds2), (H1,New_Bd_C)) :-
	subseq_body_weak_match_def( H1, Bds, A, Bds2, New_Bd_C).

recover_folded_clause( (H1,Bds), (A,Bds2), (H1,New_Bd_C)) :-
	subseq_body_match_def( H1, Bds, A, Bds2, New_Bd_C).

omit_first_unit_clauses([],[],[],[],[],[]).
omit_first_unit_clauses([C|Cls],[D|Defs],[C|Units],[D|DefUnits],Cls2,Defs2) :-
    is_a_unit_clause(D),!,
    omit_first_unit_clauses(Cls,Defs,Units,DefUnits,Cls2,Defs2).
omit_first_unit_clauses([C|Cls],[D|Defs],Units,DefUnits,[C|Cls2],[D|Defs2]) :-
    omit_first_unit_clauses(Cls,Defs,Units,DefUnits,Cls2,Defs2).


verify_recover_unit_clauses(_,[],[]).
verify_recover_unit_clauses(Folded,[U|Units],[D|DefUnits]) :-
    clause_hd_bd(D, H,bd([],[])), clause_hd_bd(U, H2,bd([],B2)),
    copy_term(Folded,Folded2), copy_term(H,H3), unfold_body(Folded2,H3, H2,B2),
    verify_recover_unit_clauses(Folded,Units,DefUnits).

unfold_body((H,Bs),A,H2,B2) :- member_rest(B,Bs,R), A = B, variant((H,R),(H2,B2)).



% sorting is performed for efficiency reason.

subseq_body_weak_match_def(ToFoldHead,ToFoldBodies,DefHead,DefBodies,NewBody) :-
	current(solver,S),ground(S),
	subseq_body_weak_match_def_(ToFoldHead,ToFoldBodies,DefHead,DefBodies, _ ,NewBody,S).

%should not be used
subseq_body_weak_match_def_(_,[],_,[],_,_,_S) :- trace, fail.


%%%subseq_body_weak_match_def_( H1,[bd([],Bd_C_unsorted)],A,[bd([],Bd_D_unsorted)],bd([],New_Bd_C2), bd([],New_Bd_C)) :- !,
%%%    same_length(Folded_Part,Bd_D_unsorted),
%%%    sort_list(Bd_D_unsorted,Bd_D), sort_list(Bd_C_unsorted,Bd_C),
%%%    free_variables(Bd_D,A,[],VarsIntD), 
%%%    folded_part(F,Folded_Part,G, Bd_C), % redo's
%%%    free_variables(Folded_Part,[],[],VarsFoldedPart),
%%%    linking_vars_folded_part(H1,F,G,VarsFoldedPart,LV),
%%%    unify_with_occurs_check(Folded_Part,Bd_D),
%%%    free_variables(Bd_D,A,[],VarsIntD2), VarsIntD==VarsIntD2,
%%%%    free_variables(VarsFoldedPart,[],[],VarsFoldedPart), % Fabio: what was this for?
%%%    free_variables(A,[],[],VA2), vars_subseteq(LV,VA2),
%%%    append([F,[A],G], New_Bd_C2),
%%%    reorder_after_sort(F,G,Bd_C,FG,InsertPos),
%%%    insert_at_position([A],InsertPos,FG,New_Bd_C).
		   
subseq_body_weak_match_def_( H1,[bd(CC,Bd_C_unsorted)|Bds],A,[bd(CD,Bd_D_unsorted)|Bds2],
			     bd3(InNewC,InEqsIneqs,New_Bd_C2), bd(NewC,New_Bd_C),S ) :-
    same_length(Folded_Part,Bd_D_unsorted),
    sort_list(Bd_D_unsorted,Bd_D), sort_list(Bd_C_unsorted,Bd_C),
    free_variables(Bd_D,A,[],VarsIntD),
    folded_part(F,Folded_Part,G, Bd_C), % redo's
    free_variables(Folded_Part,[],[],VarsFoldedPart),
    linking_vars_folded_part(H1,F,G,VarsFoldedPart,LV),
    unify_with_occurs_check(Folded_Part,Bd_D),
    free_variables(Bd_D,A,[],VarsIntD2),
    VarsIntD==VarsIntD2, % no local var has been bound
%    free_variables(VarsFoldedPart,[],[],VarsFoldedPart),  % Fabio: what was this for?
    free_variables(A,[],[],VA2),
    vars_subseteq(LV,VA2),
    append([F,[A],G], New_Bd_C2),
    %%% constraint entailment
    solve(S,CC,VarsFoldedPart,CC1),
    solve(S,CD,VarsFoldedPart,CD1),
    entails(S,CC1,CD1),

    common_constr_eqsineqs((H1,bd(CC,Bd_C)),InNewC,InEqsIneqs,S,OutNewC,OutEqsIneqs),
%%%    %%% build NewC
%%%    (var(InNewC) ->
%%%	free_variables(H1,[],[],VH1),    
%%%        solve(S,CC,VH1,OutNewC)
%%%    ;
%%%	keep_entailed(S,InNewC,CC,OutNewC)
%%%    ),
%%%    %%% build Eqs and Ineqs
%%%    body_without_equalities(Bd_C,Eqs,Bd_CwoEq),
%%%    body_without_inequalities(Bd_CwoEq,InEqs,_),
%%%    append(InEqs,Eqs,EqsIneqsC),
%%%    restrict_to_vars(EqsIneqsC,H1,EqsIneqsCH1),
%%%    (var(InEqsIneqs) ->
%%%	OutEqsIneqs = EqsIneqsCH1
%%%    ;
%%%	list_intersection(InEqsIneqs,EqsIneqsCH1,OutEqsIneqs)
%%%    ),

    %%% 
    (Bds == [] ->
	NewC = OutNewC,   write('*** newc ***:'),write(OutNewC),nl,
	Bds2 == [],
	reorder_after_sort(F,G,Bd_C,FG,InsertPos),
	insert_at_position([A],InsertPos,FG,New_Bd_C1),
	append(OutEqsIneqs,New_Bd_C1,New_Bd_C)
    ;
	subseq_body_weak_match_def_( H1,Bds,A,Bds2,bd3(OutNewC,OutEqsIneqs,New_Bd_C2), bd(NewC,New_Bd_C),S)
    ).






common_constr_eqsineqs((H1,bd(CC,Bd_C)),InNewC,InEqsIneqs,S,OutNewC,OutEqsIneqs) :-
    (var(InNewC) ->
	free_variables(H1,[],[],VH1),    
        solve(S,CC,VH1,OutNewC)
    ;
	keep_entailed(S,InNewC,CC,OutNewC)
    ),

    body_without_equalities(Bd_C,Eqs,Bd_CwoEq),
    body_without_inequalities(Bd_CwoEq,InEqs,_),
    append(InEqs,Eqs,EqsIneqsC),
    restrict_to_vars(EqsIneqsC,H1,EqsIneqsCH1),
    (var(InEqsIneqs) ->
	OutEqsIneqs = EqsIneqsCH1
    ;
	list_intersection(InEqsIneqs,EqsIneqsCH1,OutEqsIneqs)
    ).

common_constr_eqsineqs_([Cl|Cls],InNewC,InEqsIneqs,S,OutNewC,OutEqsIneqs) :-
	common_constr_eqsineqs(Cl,InNewC,InEqsIneqs,S,OutNewC1,OutEqsIneqs1),
	(Cls == [] ->
	    OutNewC = OutNewC1,
	    OutEqsIneqs = OutEqsIneqs1
	;
	    common_constr_eqsineqs_(Cls,OutNewC1,OutEqsIneqs1,S,OutNewC,OutEqsIneqs)
	).


subseq_body_match_def(_,[],_,[],_).
subseq_body_match_def( H1, [bd([],Bd_C)|Bds], A, [bd([],Bd_D)|Bds2], bd([],New_Bd_C) ) :-
    same_length(Folded_Part,Bd_D),	
    free_variables(Bd_D,A,[],VarsIntD),
    folded_part(F,Folded_Part,G, Bd_C),
    free_variables(Folded_Part,[],[],VarsFoldedPart),
    linking_vars_folded_part(H1,F,G,VarsFoldedPart,LV),
    unify_with_occurs_check(Folded_Part,Bd_D),
    free_variables(Bd_D,A,[],VarsIntD), % internal vars of Def unchanged
    free_variables(VarsFoldedPart,[],[],VarsFoldedPart), % Folded part is an instance of Bd_Def
    free_variables(A,[],[],VA2), vars_subseteq(LV,VA2),
    append([F,[A],G], New_Bd_C),
    subseq_body_match_def( H1, Bds, A, Bds2, bd([],New_Bd_C) ).



folded_part(F, Folded_Part, G, Bd_C) :-
    body_without_equalities(Bd_C, Eqs, Real_Bd_C),
    append([F, B1s, G], Real_Bd_C),
    append(Eqs, B1s, Folded_Part).

folded_part_whole_bd(Ineqs,Folded_Part,[],Bd_C) :-
    body_without_equalities(Bd_C, Eqs, Bd_C1),
    body_without_inequalities(Bd_C1,Ineqs,Bs),
    append(Eqs, Bs, Folded_Part).


linking_vars_folded_part(H,F,G,VarsFolded_Part,LV) :-
    free_variables((H,F,G),[],[],VarsHFG),
    common_vars(VarsHFG,VarsFolded_Part,LV).

%%% Added the 16/2/98 for the case F=G=[] (folding of the whole body)
linking_vars_folded_part(H,VarsFolded_Part,LV) :-
    free_variables(H,[],[],VarsHFG),
    common_vars(VarsHFG,VarsFolded_Part,LV).


reorder_after_sort(F,G,B,FG,Pos) :-
	append([F,G],FG2),
	permutation(FG2,FG),
	same_length(L,FG),
	subseq_pos(B,L,Pos),
	variant(L,FG), !.
reorder_after_sort(F,G,_,FG,1) :-
	append([F,G],FG). % in case the above fails.

subseq_pos(B,L,Pos) :-
	subseq_pos(B,L,_,1,Pos).
subseq_pos([], [], [], Pos,Pos).
subseq_pos([Head|Tail], Sbsq, [Head|Cmpl], Pos2,Pos) :-
	Pos3 is Pos2 + 1, 
	subseq_pos(Tail, Sbsq, Cmpl,Pos3,Pos).
subseq_pos([Head|Tail], [Head|Sbsq], Cmpl, Pos2,Pos) :-
	subseq_pos(Tail, Sbsq, Cmpl,Pos2,Pos).

